#include "stdafx.h"
#include <iostream>
#include <string>
#include <vector>
#include <iterator>
#include <algorithm>
#include <memory>
#include <fstream>
#include <stdexcept>
#include <boost/assert.hpp>
#include <boost/shared_ptr.hpp>
#include <boost/program_options.hpp>
#include <boost/tokenizer.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#include <harkio2/harkio.hpp>

#include "CommaSeparatedOption.hpp"
#include "Mode.hpp"
#include "OutputType.hpp"
#include "OutputFormat.hpp"
#include "validators.hpp"
#include "UsagePrinter.hpp"
#include "TemplateMaker.hpp"
#include "TFMaker.hpp"

int MIC_INDEX[1000]={0};

/** @file
 * @brief main関数などの定義。
 *
 * プログラムのエントリポイント（main関数）やヘルプメッセージを生成する関数の定義を含んでいる。
 */

/** @namespace harktool4
 * harktool4の主要機能を含む名前空間
 */

using namespace std;
namespace po = boost::program_options;

void AddOptions(po::options_description& common, po::options_description& hark_config, po::options_description& hark, po::options_description& hidden);

void AnalyzeCommandLine(int argc, char** argv, po::command_line_parser& parser, po::variables_map& vm);

void Run(const po::variables_map& vm);

void DoNothingDeleter(void*) {}

std::shared_ptr<harktool4::UsagePrinter> usage_printer;

/** @brief プログラムのエントリポイント
 */
int main(int argc, char** argv)
{
    po::options_description
        visible,
        common("Common options"),
        hark_config("Options to make HARK-config files"),
        hark("Options to make HARK files"),
        hidden("hidden options"),
        command_line_options("Command line options");

    po::positional_options_description positional_options;

    AddOptions(common, hark_config, hark, hidden);
    visible.add(common).add(hark_config).add(hark);
    usage_printer.reset(new harktool4::UsagePrinter(argv[0], visible));

    // 表示／非表示オプションをまとめる
    //command_line_options.add(common).add(hark_config).add(hark).add(hidden);
    command_line_options.add(visible).add(hidden);

    /*
     * オプション名なしで受け取れるオプション＝ポジショナルオプション
     *
     * -1を指定することで、全てのポジショナルオプションがinput-files
     * という意味になる（vector<string>で受け取れる）。
     */
    positional_options.add("input-files", -1);

    po::variables_map vm;
    po::command_line_parser parser(argc, argv);
    parser.options(command_line_options).positional(positional_options);

    AnalyzeCommandLine(argc, argv, parser, vm);
    Run(vm);

    return 0;
}

/** @brief コマンドラインオプションを定義する。
 *
 * boost::program_options::options_description へオプション定義を追加することにより、
 * コマンドラインオプションの一覧を定義する。
 * 同時にヘルプメッセージに表示するための説明文も記述している。
 */
void AddOptions(po::options_description& common, po::options_description& hark_config, po::options_description& hark, po::options_description& hidden)
{
    // オプション定義 ヘルプに表示させるオプション
    common.add_options()
        ("help", "show help")
        ("mode,m",                  po::value<harktool4::Mode>(), "what this program makes {template|tf}")
        ("output-file,o",           po::value<string>(), "output file (default is STDOUT)")
        ("output-format,f",         po::value<harktool4::OutputFormat>(), "output format {xml|binary}")
        ("output-type,t",           po::value<harktool4::OutputType>(), "output type\n(1)in template mode {mic|noise|ir|tsp}\n(2)in tf mode {m2pg|gtf}")
        ;
    hark_config.add_options()
        ("nummic",                  po::value<unsigned int>()->default_value(8), "the number of microphones")
        ("azifrom",                 po::value<double>()->default_value(0), "the start of the azimuth range in degree")
        ("aziint",                  po::value<double>()->default_value(5), "the interval of azimuth in degree")
        ("numazi",                  po::value<unsigned int>()->default_value(72), "the number of azimuth")
        ("elvfrom",                 po::value<double>()->default_value(0), "the start of the elevation range in degree")
        ("elvint",                  po::value<double>()->default_value(5), "the interval of elevation in degree")
        ("numelv",                  po::value<unsigned int>()->default_value(1), "the number of elevation")
        ("heightfrom",              po::value<double>()->default_value(0), "the start of the height range in meter")
        ("heightint",               po::value<double>()->default_value(0.01), "the interval of height in meter")
        ("numheight",               po::value<unsigned int>()->default_value(1), "the number of height")
        ("radius",                  po::value<double>()->default_value(1.0), "the radius from center to microphones or sound sources")
        ("synchronous-average",     po::value<unsigned int>()->default_value(16), "the number of synchronous averaging")
        ("original-impulse-file",   po::value<string>()->default_value("original-impulse-file.tsp"), "original impulse file")
        ("tsp-offset",              po::value<unsigned int>()->default_value(16384), "the offset of tsp signal in samples")
        ("tsp-length",              po::value<unsigned int>()->default_value(16384), "the length of one tsp in samples")
        ("signal-max",              po::value<unsigned int>()->default_value(32768), "the maximum amplitude of tsp signal")
        ("cut-start",               po::value<unsigned int>()->default_value(1), "the cut-start index in float vector")
        ("cut-end",                 po::value<unsigned int>()->default_value(0), "the cut-end index in float vector")
        ("mic-channels",            po::value<harktool4::CommaSeparatedOption<unsigned int>>(), "a comma separated list of the microphone channels")
        ("filepath",                po::value<string>(), "the file path template string used by ir and tsp")
        ;
    hark.add_options()
        ("nfft",                    po::value<unsigned int>()->default_value(512), "the number of FFT points")
        ("normalize-src",           "normalization of sound sources axis")
        ("normalize-mic",           "normalization of microphone channels axis")
        ("normalize-freq",          "normalization of frequency axis")
        ("geometry-calculus",       po::value<unsigned int>()->default_value(0), "geometry calculus")
        ("sampling-freq",           po::value<unsigned int>()->default_value(16000), "sampling frequency of TSP signal or Impulse Response")
        ("direct-length",           po::value<unsigned int>()->default_value(32), "direct sound length")
        ;
    // オプション定義 非表示のオプション
    hidden.add_options()
        ("input-files", po::value<vector<string>>(), "input files for making a transfer function")
        ;
}

/** @brief コマンドライン引数を解析する。
 */
void AnalyzeCommandLine(int argc, char** argv, po::command_line_parser& parser, po::variables_map& vm)
{
    try
    {
        // コマンドラインを解析
        store(parser.run(), vm);
        notify(vm);
    }
    catch (const po::invalid_option_value& e)
    {
        // 解析中にエラー
        cerr << e.what() << endl;
        exit(1);
    }
    catch (const po::multiple_occurrences& e)
    {
        // 1回しか指定できないオプションを複数回指定した

        // multiple_occurrences::get_option_nameはboost 1.40には存在しない
        // cerr << e.get_option_name() << " can't be specified more than one time" << endl;
        cerr << e.what() << endl;
        exit(1);
    }
    catch (const po::unknown_option& e)
    {
        cerr << e.what() << endl;
        usage_printer->Print(cerr);
        exit(1);
    }
    catch (const std::exception& e)
    {
        cerr << e.what() << endl;
        exit(1);
    }
}

/** @brief 指定されたオプションに応じた処理を行う。
 *
 * @param vm 処理内容を指定するオプション群
 */
void Run(const po::variables_map& vm)
{
    if (vm.count("help"))
    {
        usage_printer->Print(cout);
        return;
    }

    if (vm.count("mode") == 0)
    {
        std::cout << "You may want to use --mode to specify creation mode\n";
        return;
    }

    std::shared_ptr<std::ostream> output_file(&cout, DoNothingDeleter);
    if (vm.count("output-file"))
    {
        output_file.reset(new std::ofstream(vm["output-file"].as<string>()));
    }

    auto mode = vm["mode"].as<harktool4::Mode>();
    switch (mode)
    {
    case harktool4::Mode::Template:
        harktool4::TemplateMaker(vm, output_file).Run();
        break;
    case harktool4::Mode::TF:
        harktool4::TFMaker(vm, output_file).Run();
        break;
    }
}

